# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from inspect import Parameter
from typing import Any, Dict, List, Optional, TypeVar

from .dependency_injection import DependencyResolver
from .errors import (
    HeaderMissing, HTTPError, ParamMissing, RequestParserNotAvailable, ValidationError
)
from .http import HTTP_400, Cookies, Headers, QueryParams
from .parsers import RequestParser
from .router import Route
from .typing import (
    Header, QueryParam, RequestBody, RequestData, RequestInput, extract_optional_annotation
)
from .validation import is_schema, load_schema

_T = TypeVar("_T")


class HeaderComponent:
    """Retrieves a named header from the request.

    Examples:

      def handle(content_type: Header) -> Response:
        ...

      def handle(content_type: Optional[Header]) -> Response:
        ...
    """

    is_cacheable = False
    is_singleton = False

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        _, annotation = extract_optional_annotation(parameter.annotation)
        return annotation is Header

    def resolve(self, parameter: Parameter, headers: Headers) -> Optional[str]:
        is_optional, _ = extract_optional_annotation(parameter.annotation)
        header_name = parameter.name.replace("_", "-")

        try:
            return headers[header_name]
        except HeaderMissing:
            if is_optional:
                return None

            raise HTTPError(HTTP_400, {"errors": {header_name: "missing"}})


class QueryParamComponent:
    """Retrieves a named query param from the request.

    Examples:

      def handle(x: QueryParam) -> Response:
        ...

      def handle(x: Optional[QueryParam]) -> Response:
        ...
    """

    is_cacheable = False
    is_singleton = False

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        _, annotation = extract_optional_annotation(parameter.annotation)
        return annotation is QueryParam

    def resolve(self, parameter: Parameter, params: QueryParams) -> Optional[str]:
        is_optional, _ = extract_optional_annotation(parameter.annotation)

        try:
            return params[parameter.name]
        except ParamMissing:
            if is_optional:
                return None

            raise HTTPError(HTTP_400, {"errors": {parameter.name: "missing"}})


class RequestBodyComponent:
    """A component that reads the entire request body into a string.

    Examples:

      def handle(body: RequestBody) -> Response:
        ...
    """

    is_cacheable = True
    is_singleton = False

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        return parameter.annotation is RequestBody

    def resolve(self, content_length: Header, body_file: RequestInput) -> RequestBody:
        return RequestBody(body_file.read(int(content_length)))


class RequestDataComponent:
    """A component that parses request data based on the content-type
    header and the set of registered request parsers.  If no request
    parser is available, then an HTTP 415 response is returned.

    Examples:

      def handle(data: RequestData) -> Response:
        ...
    """

    __slots__ = ["parsers"]

    is_cacheable = True
    is_singleton = False

    def __init__(self, parsers: List[RequestParser]) -> None:
        self.parsers = parsers

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        return parameter.annotation is RequestData

    def resolve(self, content_type: Optional[Header], resolver: DependencyResolver) -> RequestData:
        content_type_str = (content_type or "").lower()
        for parser in self.parsers:
            if parser.can_parse_content(content_type_str):
                return RequestData(resolver.resolve(parser.parse)())
        raise RequestParserNotAvailable(content_type_str)


class CookiesComponent:
    """A component that parses request cookies.

    Examples:

      def handle(cookies: Cookies) -> Response:
        cookies["some-cookie"]
    """

    is_cacheable = True
    is_singleton = False

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        return parameter.annotation is Cookies

    def resolve(self, cookie: Optional[Header]) -> Cookies:
        if cookie is None:
            return Cookies()
        return Cookies.parse(cookie)


class RouteComponent:
    """A component that resolves the current route.
    """

    __slots__ = ["route"]

    is_cacheable = True
    is_singleton = False

    def __init__(self, route: Optional[Route]) -> None:
        self.route = route

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        _, annotation = extract_optional_annotation(parameter.annotation)
        return annotation is Route

    def resolve(self) -> Optional[Route]:
        return self.route


class RouteParamsComponent:
    """A component that resolves route params.

    Examples:

      def handle(name: str, age: int) -> Response:
        ...

      app.add_route(Route("/{name}/{age}", handle))
    """

    __slots__ = ["params"]

    is_cacheable = False
    is_singleton = False

    def __init__(self, params: Dict[str, str]) -> None:
        self.params = params

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        return parameter.name in self.params

    def resolve(self, parameter: Parameter) -> Any:
        try:
            return parameter.annotation(self.params[parameter.name])
        except (TypeError, ValueError):
            raise HTTPError(HTTP_400, {"errors": {
                parameter.name: f"invalid {parameter.annotation.__name__} value",
            }})


class SchemaComponent:
    """A component that validates request data according to a schema.
    """

    is_cacheable = False
    is_singleton = False

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        return is_schema(parameter.annotation)

    def resolve(self, parameter: Parameter, data: RequestData) -> Any:
        try:
            return load_schema(parameter.annotation, data)
        except ValidationError as e:
            raise HTTPError(HTTP_400, {"errors": e.reasons})



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
from io import BytesIO
from typing import Any, Callable

from molten import HTTP_200, Request, Response

try:
    from prometheus_client import (
        CONTENT_TYPE_LATEST, CollectorRegistry, Counter, Gauge, Histogram,
        generate_latest, multiprocess
    )
except ImportError:  # pragma: no cover
    raise ImportError("'prometheus_client' package missing. Run 'pip install prometheus-client'.")

REQUEST_DURATION = Histogram(
    "http_request_duration_seconds",
    "Time spent processing a request.",
    ["method", "path"],
)
REQUEST_COUNT = Counter(
    "http_requests_total",
    "Request count by method, path and status line.",
    ["method", "path", "status"],
)
REQUESTS_INPROGRESS = Gauge(
    "http_requests_inprogress",
    "Requests in progress by method and path",
    ["method", "path"],
)

#: Micro-optimization to avoid allocating a new dict on every metrics
#: request.  Response itself copies the headers its given so this
#: shouldn't be a problem.
_HEADERS = {"content-type": CONTENT_TYPE_LATEST}


def expose_metrics() -> Response:
    """Expose prometheus metrics from the current process.
    """
    return Response(HTTP_200, headers=_HEADERS, stream=BytesIO(generate_latest()))


def expose_metrics_multiprocess() -> Response:  # pragma: no cover
    """Expose prometheus metrics from the current set of processes.
    Use this instead of expose_metrics if you're using a multi-process
    server.
    """
    registry = CollectorRegistry()
    multiprocess.MultiProcessCollector(registry)
    return Response(HTTP_200, headers=_HEADERS, stream=BytesIO(generate_latest(registry)))


def prometheus_middleware(handler: Callable[..., Any]) -> Callable[..., Any]:
    """Collect prometheus metrics from your handlers.
    """

    def middleware(request: Request) -> Any:
        status = "500 Internal Server Error"
        start_time = time.monotonic()
        requests_inprogress = REQUESTS_INPROGRESS.labels(request.method, request.path)
        requests_inprogress.inc()

        try:
            response = handler()
            status = response.status
            return response
        finally:
            requests_inprogress.dec()
            REQUEST_COUNT.labels(request.method, request.path, status).inc()
            REQUEST_DURATION.labels(request.method, request.path).observe(time.monotonic() - start_time)
    return middleware



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Type


class _Missing:
    """The type of missing values.
    """

    def __repr__(self) -> str:  # pragma: no cover
        return "Missing"


#: Canary value representing missing attributes or values.
Missing = _Missing()


def is_schema(ob: Type[Any]) -> bool:
    """Returns True if the given type is a schema.
    """
    return isinstance(ob, type) and \
        hasattr(ob, "_SCHEMA") and \
        hasattr(ob, "_FIELDS")



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .cookies import Cookie, Cookies
from .headers import Headers
from .query_params import QueryParams
from .request import Request
from .response import Response, StreamingResponse
from .status_codes import *
from .uploaded_file import UploadedFile

__all__ = [
    "Request", "Response", "StreamingResponse",
    "Headers", "QueryParams", "Cookies", "Cookie", "UploadedFile",

    # 1xx
    "HTTP_100", "HTTP_101", "HTTP_102",

    # 2xx
    "HTTP_200", "HTTP_201", "HTTP_202", "HTTP_203", "HTTP_204", "HTTP_205", "HTTP_206", "HTTP_207", "HTTP_208",

    # 3xx
    "HTTP_300", "HTTP_301", "HTTP_302", "HTTP_303", "HTTP_304", "HTTP_305", "HTTP_307", "HTTP_308",

    # 4xx
    "HTTP_400", "HTTP_401", "HTTP_402", "HTTP_403", "HTTP_404", "HTTP_405", "HTTP_406", "HTTP_407", "HTTP_408",
    "HTTP_409", "HTTP_410", "HTTP_411", "HTTP_412", "HTTP_413", "HTTP_414", "HTTP_415", "HTTP_416", "HTTP_417",
    "HTTP_418", "HTTP_421", "HTTP_422", "HTTP_423", "HTTP_424", "HTTP_426", "HTTP_428", "HTTP_429", "HTTP_431",
    "HTTP_444", "HTTP_451", "HTTP_499",

    # 5xx
    "HTTP_500", "HTTP_501", "HTTP_502", "HTTP_503", "HTTP_504", "HTTP_505", "HTTP_506", "HTTP_507", "HTTP_508",
    "HTTP_510", "HTTP_511", "HTTP_599",
]



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict, List, Optional

import pkg_resources

from ..app import BaseApp
from ..http import HTTP_200, Request, Response
from .documents import Metadata, SecurityScheme, generate_openapi_document


class OpenAPIHandler:
    """Dynamically generates and serves OpenAPI v3 documents based on
    the current application object.  Once generated, the document is
    subsequently served from cache.

    Examples:

      >>> get_schema = OpenAPIHandler(
      ...   metadata=Metadata(title="Pet Store", version="0.1.0"),
      ...   security_schemes=[HTTPSecurityScheme("Bearer", "bearer")],
      ...   default_security_scheme="Bearer",
      ... )

    Parameters:
      metadata: Various meta-information about the current API.
      security_schemes: A list of security schemes used throughout the
        API.
    """

    def __init__(
            self,
            metadata: Metadata,
            security_schemes: Optional[List[SecurityScheme]] = None,
            default_security_scheme: Optional[str] = None,
    ) -> None:
        self.metadata = metadata
        self.security_schemes = security_schemes or []
        self.default_security_scheme = default_security_scheme
        self.document: Optional[Dict[str, Any]] = None

    @property
    def __name__(self) -> str:
        return type(self).__name__  # type: ignore

    def __call__(self, app: BaseApp) -> Optional[Dict[str, Any]]:
        """Generates an OpenAPI v3 document.
        """
        if not self.document:
            self.document = generate_openapi_document(
                app,
                self.metadata,
                self.security_schemes,
                self.default_security_scheme,
            )

        return self.document


class OpenAPIUIHandler:
    """Renders the `Swagger UI`_.

    Parameters:
      schema_route_name: The name of the route that exposes the
        schema.  The actual path to the schema is looked up whenever
        the handler is called.

    .. _Swagger UI: https://github.com/swagger-api/swagger-ui
    """

    def __init__(self, schema_route_name: str = "OpenAPIHandler") -> None:
        self.schema_route_name = schema_route_name
        self.template = pkg_resources.resource_string(f"molten.openapi.templates", "index.html").decode("utf-8")

    @property
    def __name__(self) -> str:
        return type(self).__name__  # type: ignore

    def __call__(self, app: BaseApp, request: Request) -> Response:
        """Renders the Swagger UI.
        """
        schema_uri = app.reverse_uri(self.schema_route_name)
        rendered_template = self.template % {"schema_uri": schema_uri}
        return Response(HTTP_200, content=rendered_template, headers={
            "content-type": "text/html",
        })



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
from typing import Dict, Iterable, Iterator, List, Optional, Tuple, Union

from ..errors import HeaderMissing
from ..typing import Environ

#: An alias representing a dictionary of headers.
HeadersDict = Dict[str, Union[str, List[str]]]

#: WSGI keeps these separate from other headers.
CONTENT_VARS = {"CONTENT_LENGTH", "CONTENT_TYPE"}


class Headers(Iterable[Tuple[str, str]]):
    """A mapping from case-insensitive header names to lists of values.
    """

    __slots__ = ["_headers"]

    def __init__(self, mapping: Optional[HeadersDict] = None) -> None:
        self._headers: Dict[str, List[str]] = defaultdict(list)
        self.add_all(mapping or {})

    @classmethod
    def from_environ(cls, environ: Environ) -> "Headers":
        """Construct a Headers instance from a WSGI environ.
        """
        headers = {}
        for name, value in environ.items():
            if name in CONTENT_VARS:
                headers[name.replace("_", "-")] = value

            elif name.startswith("HTTP_"):
                headers[_parse_environ_header(name)] = [value]

        return cls(headers)

    def add(self, header: str, value: Union[str, List[str]]) -> None:
        """Add values for a particular header.
        """
        if isinstance(value, list):
            self._headers[header.lower()].extend(value)
        else:
            self._headers[header.lower()].append(value)

    def add_all(self, mapping: HeadersDict) -> None:
        """Add a group of headers.
        """
        for header, value_or_values in mapping.items():
            self.add(header, value_or_values)

    def get(self, header: str, default: Optional[str] = None) -> Optional[str]:
        """Get the last value for a given header.
        """
        try:
            return self[header]
        except HeaderMissing:
            return default

    def get_all(self, header: str) -> List[str]:
        """Get all the values for a given header.
        """
        return self._headers[header.lower()]

    def get_int(self, header: str, default: Optional[int] = None) -> Optional[int]:
        """Get the last value for a given header as an integer.
        """
        try:
            return int(self[header])
        except HeaderMissing:
            return default

    def __delitem__(self, header: str) -> None:
        """Delete all the values for a given header.
        """
        del self._headers[header.lower()]

    def __getitem__(self, header: str) -> str:
        """Get the last value for a given header.

        Raises:
          HeaderMissing: When the header is missing.
        """
        try:
            return self._headers[header.lower()][-1]
        except IndexError:
            raise HeaderMissing(header)

    def __setitem__(self, header: str, value: str) -> None:
        """Replace a header's values.
        """
        self._headers[header.lower()] = [value]

    def __iter__(self) -> Iterator[Tuple[str, str]]:
        """Iterate over all the headers.
        """
        for header, values in self._headers.items():
            for value in values:
                yield header, value

    def __repr__(self) -> str:
        mapping = ", ".join(f"{repr(name)}: {repr(value)}" for name, value in self._headers.items())
        return f"Headers({{{mapping}}})"


#: The number of characters that are stripped from the beginning of
#: every header name in a WSGI environ.
HEADER_PREFIX_LEN = len("HTTP_")

#: A lookup table from WSGI header strings to header names.
HEADER_PARSER_CACHE: Dict[str, str] = {}


def _parse_environ_header(header: str) -> str:
    try:
        return HEADER_PARSER_CACHE[header]
    except KeyError:
        HEADER_PARSER_CACHE[header] = parsed_header = header[HEADER_PREFIX_LEN:].replace("_", "-")
        return parsed_header



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import inspect
from inspect import Parameter
from typing import Any, Callable, Dict, Optional, Sequence, no_type_check

from molten import BaseApp

try:
    import dramatiq
except ImportError:  # pragma: no cover
    raise ImportError("'dramatiq' package missing. Run 'pip install dramatiq'.")


#: The global dependency injector instace.  Call setup_dramatiq to set
#: this up.
_INJECTOR = None


def setup_dramatiq(app: BaseApp) -> None:
    """Sets up the global state required to be able to inject
    components into Dramatiq actors.

    Examples:

      >>> from molten.contrib.dramatiq import setup_dramatiq

      >>> # All components that were registered with your app will be
      >>> # available to your actors once you call this function.
      >>> setup_dramatiq(app)

    """
    global _INJECTOR
    _INJECTOR = app.injector


@no_type_check
def actor(fn=None, **kwargs):
    """Use this in place of dramatiq.actor in order to create actors
    that can request components via dependency injection.  This is
    just a wrapper around dramatiq.actor and it takes the same
    set of parameters.

    Examples:

      >>> from molten.contrib.dramatiq import actor

      >>> @actor(queue_name="example")
      ... def add(x, y, database: Database) -> None:
      ...   database.put(x + y)
      ...
      >>> add.send(1, 2)

    """
    def decorator(fn):
        return dramatiq.actor(_inject(fn), **kwargs)

    if fn is None:
        return decorator
    return decorator(fn)


@no_type_check
def _inject(fn: Optional[Callable[..., Any]] = None) -> Callable[..., Any]:
    def decorator(fn):
        parameters = {name: i for i, name in enumerate(inspect.signature(fn).parameters)}

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            try:
                resolver = _INJECTOR.get_resolver()
                resolver.add_component(_ArgumentResolver(parameters, args, kwargs))
            except AttributeError:  # pragma: no cover
                raise RuntimeError(
                    "Dramatiq support is not set up correctly. "
                    "Don't forget to call setup_dramatiq()."
                )

            resolved_fn = resolver.resolve(fn)
            return resolved_fn()

        return wrapper

    if fn is None:  # pragma: no cover
        return decorator
    return decorator(fn)


class _ArgumentResolver:
    is_cacheable = False
    is_singleton = False

    def __init__(self, parameters: Dict[str, int], args: Sequence[Any], kwargs: Dict[str, Any]) -> None:
        self.state = state = kwargs
        for name, idx in parameters.items():
            if name not in state:
                try:
                    state[name] = args[idx]
                except IndexError:
                    continue

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        return parameter.name in self.state or \
            parameter.default is not Parameter.empty

    def resolve(self, parameter: Parameter) -> Any:
        try:
            return self.state[parameter.name]
        except KeyError:
            return parameter.default



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
from functools import partial
from io import BytesIO
from json import dumps as to_json
from typing import Any, BinaryIO, Callable, Dict, Optional, Union, cast
from urllib.parse import urlencode

from ..app import BaseApp
from ..http import HTTP_200
from ..http.headers import Headers, HeadersDict
from ..http.query_params import ParamsDict, QueryParams
from ..http.request import Request
from ..http.response import Response
from ..typing import Environ
from .common import to_environ

try:
    from gunicorn.http.errors import NoMoreData  # type: ignore
except ImportError:  # pragma: no cover
    class NoMoreData(Exception):  # type: ignore
        pass


HTTP_METHODS = {"delete", "head", "get", "options", "patch", "post", "put"}


class TestResponse:
    """A wrapper around Response objects that adds a few additional
    helper methods for testing.

    Attributes:
      status: The response status line.
      status_code: The response status code as an integer.
      headers: The response headers.
      stream: The response data as a binary file.
      data: The response data as a string.
    """

    __slots__ = ["_response"]

    def __init__(self, response: Response) -> None:
        self._response = response

    @property
    def data(self) -> str:
        """Rewinds the output stream and returns all its data.
        """
        self._response.stream.seek(0)
        return self._response.stream.read().decode("utf-8")

    @property
    def status_code(self) -> int:
        """Returns the HTTP status code as an integer.
        """
        code, _, _ = self._response.status.partition(" ")
        return int(code)

    def json(self) -> Any:
        """Convert the response data to JSON.
        """
        return json.loads(self.data)

    def __getattr__(self, name: str) -> Any:
        return getattr(self._response, name)


class TestClient:
    """Test clients are used to simulate requests against an
    application instance.
    """

    __slots__ = ["app"]

    def __init__(self, app: BaseApp) -> None:
        self.app = app

    def request(
            self,
            method: str,
            path: str,
            headers: Optional[Union[HeadersDict, Headers]] = None,
            params: Optional[Union[ParamsDict, QueryParams]] = None,
            body: Optional[bytes] = None,
            data: Optional[Dict[str, str]] = None,
            json: Optional[Any] = None,
            auth: Optional[Callable[[Request], Request]] = None,
            prepare_environ: Optional[Callable[[Environ], Environ]] = None,
    ) -> TestResponse:
        """Simulate a request against the application.

        Raises:
          RuntimeError: If both 'data' and 'json' are provided.

        Parameters:
          method: The request method.
          path: The request path.
          headers: Optional request headers.
          params: Optional query params.
          body: An optional bytestring for the request body.
          data: An optional dictionary for the request body that gets url-encoded.
          json: An optional value for the request body that gets json-encoded.
          auth: An optional function that can be used to add auth
            headers to the request.
        """
        if data is not None and json is not None:
            raise RuntimeError("either 'data' or 'json' should be provided, not both")

        request = Request(
            method=method.upper(),
            path=path,
            headers=headers,
            params=params,
        )
        if body is not None:
            request.headers["content-length"] = f"{len(body)}"
            request.body_file = BytesIO(body)

        if data is not None:
            request_content = urlencode(data).encode("utf-8")
            request.headers["content-type"] = "application/x-www-form-urlencoded"
            request.headers["content-length"] = f"{len(request_content)}"
            request.body_file = BytesIO(request_content)

        elif json is not None:
            request_content = to_json(json).encode("utf-8")
            request.headers["content-type"] = "application/json; charset=utf-8"
            request.headers["content-length"] = f"{len(request_content)}"
            request.body_file = BytesIO(request_content)

        if auth is not None:
            request = auth(request)

        response = Response(HTTP_200)

        def start_response(status, response_headers, exc_info=None):  # type: ignore
            nonlocal response
            response.status = status
            response.headers = Headers(dict(response_headers))

        try:
            environ = to_environ(request)
            if prepare_environ:  # pragma: no cover
                environ = prepare_environ(environ)

            chunks = self.app(environ, start_response)
        except NoMoreData:
            chunks = []

        if response.headers.get("transfer-encoding") == "chunked":
            response.stream = cast(BinaryIO, chunks)
            return TestResponse(response)

        for chunk in chunks:
            response.stream.write(chunk)

        response.stream.seek(0)
        return TestResponse(response)

    def __getattr__(self, name: str) -> Any:
        if name in HTTP_METHODS:
            return partial(self.request, name)
        raise AttributeError(f"unknown attribute {name}")



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from io import BytesIO
from typing import BinaryIO, Dict, List, Optional, Union

from ..typing import Environ
from .headers import Headers
from .query_params import ParamsDict, QueryParams


class Request:
    """Represents an individual HTTP request.

    Attributes:
      method: The HTTP method.
      scheme: The URL scheme.
      host: The hostname.
      port: The port.
      path: The path.
      params: The query parameters.
      headers: The request headers.
      body_file: A file-like object representing the request body.
    """

    __slots__ = [
        "method",
        "scheme",
        "host",
        "port",
        "path",
        "params",
        "headers",
        "body_file",
    ]

    def __init__(
            self, *,
            method: str = "GET",
            scheme: str = "http",
            host: str = "127.0.0.1",
            port: int = 8000,
            path: str = "/",
            params: Optional[Union[ParamsDict, QueryParams]] = None,
            headers: Optional[Union[Dict[str, Union[str, List[str]]], Headers]] = None,
            body_file: Optional[BinaryIO] = None,
    ) -> None:
        self.method = method
        self.scheme = scheme
        self.host = host
        self.port = port
        self.path = path
        self.body_file = body_file or BytesIO()

        if isinstance(headers, dict):
            self.headers: Headers = Headers(headers)
        else:
            self.headers = headers or Headers()

        if not params:
            self.params = QueryParams()
        elif isinstance(params, dict):
            self.params = QueryParams(params)
        else:
            self.params = params

    @classmethod
    def from_environ(cls, environ: Environ) -> "Request":
        """Construct a Request object from a WSGI environ.
        """
        return Request(
            method=environ["REQUEST_METHOD"],
            scheme=environ["wsgi.url_scheme"],
            host=environ.get("HTTP_HOST", ""),
            port=environ.get("SERVER_PORT", 0),
            path=environ.get("PATH_INFO", ""),
            params=QueryParams.from_environ(environ),
            headers=Headers.from_environ(environ),
            body_file=environ["wsgi.input"],
        )

    def __repr__(self) -> str:
        return (
            f"Request(method={repr(self.method)}, scheme={repr(self.scheme)}, host={repr(self.host)}, "
            f"port={repr(self.port)}, path={repr(self.path)}, params={repr(self.params)}, "
            f"headers={repr(self.headers)}, body_file={repr(self.body_file)})"
        )



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple
from inspect import Parameter
from typing import Any, Callable, NewType, Optional

from molten import DependencyResolver, Settings

try:
    from sqlalchemy import create_engine  # type: ignore
    from sqlalchemy.orm import Session, sessionmaker
except ImportError:  # pragma: no cover
    raise ImportError("'sqlalchemy' package missing. Run 'pip install sqlalchemy'.")


#: The type of session factories.
SessionFactory = NewType("SessionFactory", sessionmaker)  # type: ignore

#: A named tuple containing an instantiated SQLAlchemy ``engine``
#: object and the ``session_factory``.
EngineData = namedtuple("EngineData", "engine, session_factory")


class SQLAlchemyEngineComponent:
    """A component that sets up an SQLAlchemy Engine.  This component
    depends on the availability of a :class:`molten.Settings`
    component.

    Your settings dictionary must contain a ``database_engine_dsn``
    setting pointing at the database to use.  Additionally, you may
    provide a ``database_engine_params`` setting represeting
    dictionary data that will be passed directly to
    ``sqlalchemy.create_engine``.

    Examples:

      >>> from molten import App
      >>> from molten.contrib.sqlalchemy import SQLAlchemyEngineComponent, SQLAlchemySessionComponent, SQLAlchemyMiddleware
      >>> from molten.contrib.toml_settings import TOMLSettingsComponent

      >>> app = App(
      ...   components=[
      ...     TOMLSettingsComponent(),
      ...     SQLAlchemyEngineComponent(),
      ...     SQLAlchemySessionComponent(),
      ...   ],
      ...   middleware=[SQLAlchemyMiddleware()],
      ... )
    """

    is_cacheable = True
    is_singleton = True

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        return parameter.annotation is EngineData

    def resolve(self, settings: Settings) -> EngineData:
        engine = create_engine(
            settings.strict_get("database_engine_dsn"),
            **settings.get("database_engine_params", {}),
        )

        session_factory = sessionmaker()
        session_factory.configure(bind=engine)
        return EngineData(engine, session_factory)


class SQLAlchemySessionComponent:
    """A component that creates and injects SQLAlchemy sessions.

    Examples:

      >>> def find_todos(session: Session) -> List[Todo]:
      ...   todos = session.query(TodoModel).all()
      ...   ...

    """

    is_cacheable = True
    is_singleton = False

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        return parameter.annotation is Session

    def resolve(self, engine_data: EngineData) -> Session:  # type: ignore
        return engine_data.session_factory()


class SQLAlchemyMiddleware:
    """A middleware that automatically commits SQLAlchemy sessions on
    handler success and automatically rolls back sessions on handler
    failure.

    Sessions are only instantiated and operated upon if the handler or
    any other middleware has requested an SQLAlchemy session object
    via DI.  This means that handlers that don't request a Session
    object don't automatically connect to the Database.
    """

    def __call__(self, handler: Callable[..., Any]) -> Callable[..., Any]:
        def middleware(resolver: DependencyResolver) -> Any:
            session = None

            try:
                response = handler()
                session = get_optional_session(resolver)
                if session is not None:
                    session.commit()

                return response
            except Exception:
                session = get_optional_session(resolver)
                if session is not None:
                    session.rollback()

                raise
            finally:
                if session is not None:
                    session.close()

        return middleware


def get_optional_session(resolver: DependencyResolver) -> Optional[Session]:  # type: ignore
    """Get a session object from the resolver iff one was previously
    requested.  Returns None if no function has requested a session so
    far.
    """
    for component, value in resolver.instances.items():
        if type(component) is SQLAlchemySessionComponent:
            return value
    return None



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# 1xx
HTTP_100 = "100 Continue"
HTTP_101 = "101 Switching Protocols"
HTTP_102 = "102 Processing"

# 2xx
HTTP_200 = "200 OK"
HTTP_201 = "201 Created"
HTTP_202 = "202 Accepted"
HTTP_203 = "203 Non-Authoritative Information"
HTTP_204 = "204 No Content"
HTTP_205 = "205 Reset Content"
HTTP_206 = "206 Partial Content"
HTTP_207 = "207 Multi-Status"
HTTP_208 = "208 Already Reported"

# 3xx
HTTP_300 = "300 Multiple Choices"
HTTP_301 = "301 Moved Permanently"
HTTP_302 = "302 Found"
HTTP_303 = "303 See Other"
HTTP_304 = "304 Not Modified"
HTTP_305 = "305 Use Proxy"
HTTP_307 = "307 Temporary Redirect"
HTTP_308 = "308 Permanent Redirect"

# 4xx
HTTP_400 = "400 Bad Request"
HTTP_401 = "401 Unauthorized"
HTTP_402 = "402 Payment Required"
HTTP_403 = "403 Forbidden"
HTTP_404 = "404 Not Found"
HTTP_405 = "405 Method Not Allowed"
HTTP_406 = "406 Not Acceptable"
HTTP_407 = "407 Proxy Authentication Required"
HTTP_408 = "408 Request Timeout"
HTTP_409 = "409 Conflict"
HTTP_410 = "410 Gone"
HTTP_411 = "411 Length Required"
HTTP_412 = "412 Precondition Failed"
HTTP_413 = "413 Payload Too Large"
HTTP_414 = "414 Request-URI Too Long"
HTTP_415 = "415 Unsupported Media Type"
HTTP_416 = "416 Requested Range Not Satisfiable"
HTTP_417 = "417 Expectation Failed"
HTTP_418 = "418 I'm a teapot"
HTTP_421 = "421 Misdirected Request"
HTTP_422 = "422 Unprocessable Entity"
HTTP_423 = "423 Locked"
HTTP_424 = "424 Failed Dependency"
HTTP_426 = "426 Upgrade Required"
HTTP_428 = "428 Precondition Required"
HTTP_429 = "429 Too Many Requests"
HTTP_431 = "431 Request Header Fields Too Large"
HTTP_444 = "444 Connection Closed Without Response"
HTTP_451 = "451 Unavailable For Legal Reasons"
HTTP_499 = "499 Client Closed Request"
HTTP_415 = "415 Unsupported Media Type"

# 5xx
HTTP_500 = "500 Internal Server Error"
HTTP_501 = "501 Not Implmeneted"
HTTP_502 = "502 Bad Gateway"
HTTP_503 = "503 Service Unavailable"
HTTP_504 = "504 Gateway Timeout"
HTTP_505 = "505 HTTP Version Not Supported"
HTTP_506 = "506 Variant Also Negotiates"
HTTP_507 = "507 Insufficient Storage"
HTTP_508 = "508 Loop Detected"
HTTP_510 = "510 Not Extended"
HTTP_511 = "511 Network Authentication Required"
HTTP_599 = "599 Network Connect Timeout Error"



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from shutil import copyfileobj
from typing import Any, BinaryIO, Union

from .headers import Headers


class UploadedFile:
    """Represents a file that was uploaded as part of an HTTP request.
    May be backed by an in-memory file-like object or a real temporary
    file on disk.

    Attributes:
      filename: The name the file had in the request.
      headers: Headers sent with the file.
      stream: The file-like object containing the data.
    """

    __slots__ = [
        "filename",
        "headers",
        "stream",
    ]

    def __init__(self, filename: str, headers: Headers, stream: BinaryIO) -> None:
        self.filename = filename
        self.headers = headers
        self.stream = stream

    def save(self, destination: Union[str, BinaryIO]) -> None:
        """Save the file's contents either to another file object or to a path on disk.
        """
        if isinstance(destination, str):
            with open(destination, "wb+") as outfile:
                copyfileobj(self.stream, outfile)

        else:
            copyfileobj(self.stream, destination)

    def __getattr__(self, name: str) -> Any:
        return getattr(self.stream, name)

    def __repr__(self) -> str:
        params = ", ".join(f"{name}={getattr(self, name)!r}" for name in self.__slots__)
        return f"UploadedFile({params})"



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Tuple, TypeVar, Union

KT = TypeVar("KT")
VT = TypeVar("VT")

Mapping = Union[
    Dict[KT, Union[VT, List[VT]]],
    Iterable[Tuple[KT, Union[VT, List[VT]]]]
]


class MultiDict(Iterable[Tuple[KT, VT]]):
    """A mapping from param names to lists of values.  Once
    constructed, these instances cannot be modified.
    """

    __slots__ = ["_data"]

    def __init__(self, mapping: Optional[Mapping[KT, VT]] = None) -> None:
        self._data: Dict[KT, List[VT]] = defaultdict(list)
        self._add_all(mapping or {})

    def _add(self, name: KT, value: Union[VT, List[VT]]) -> None:
        """Add values for a particular key.
        """
        if isinstance(value, list):
            self._data[name].extend(value)
        else:
            self._data[name].append(value)

    def _add_all(self, mapping: Mapping[KT, VT]) -> None:
        """Add a group of values.
        """
        items: Iterable[Tuple[KT, Union[VT, List[VT]]]]

        if isinstance(mapping, dict):
            items = mapping.items()
        else:
            items = mapping

        for name, value_or_values in items:
            self._add(name, value_or_values)

    def get(self, name: KT, default: Optional[VT] = None) -> Optional[VT]:
        """Get the last value for a given key.
        """
        try:
            return self[name]
        except KeyError:
            return default

    def get_all(self, name: KT) -> List[VT]:
        """Get all the values for a given key.
        """
        return self._data[name]

    def __getitem__(self, name: KT) -> VT:
        """Get the last value for a given key.

        Raises:
          KeyError: When the key is missing.
        """
        try:
            return self._data[name][-1]
        except IndexError:
            raise KeyError(name)

    def __iter__(self) -> Iterator[Tuple[KT, VT]]:
        """Iterate over all the parameters.
        """
        for name, values in self._data.items():
            for value in values:
                yield name, value

    def __repr__(self) -> str:
        mapping = ", ".join(f"{repr(name)}: {repr(value)}" for name, value in self._data.items())
        return f"{type(self).__name__}({{{mapping}}})"


def annotate(**options: Any) -> Callable[..., Any]:
    """Add arbitrary attributes to a callable.

    Examples:

      >>> @annotate(openapi_tags=["a", "b"])
      ... def some_handler():
      ...   ...

      >>> some_handler.openapi_tags
      ["a", "b"]
    """
    def wrapper(fn: Callable[..., Any]) -> Callable[..., Any]:
        for name, value in options.items():
            setattr(fn, name, value)

        return fn
    return wrapper



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, BinaryIO, Callable, Dict, List, NewType, Tuple, Union

#: An alias representing a WSGI environment dictionary.
Environ = Dict[str, Any]

#: An alias representing a WSGI start_response callback.
StartResponse = Callable[[str, List[Tuple[str, str]], Any], None]

#: The type of middleware functions.
Middleware = Callable[[Callable[..., Any]], Callable[..., Any]]

#: The request method.
Method = NewType("Method", str)

#: The request URI scheme.
Scheme = NewType("Scheme", str)

#: The request hostname.
Host = NewType("Host", str)

#: The request port.
Port = NewType("Port", int)

#: The request query string.
QueryString = NewType("QueryString", str)

#: A query string parameter.
QueryParam = NewType("QueryParam", str)

#: A header.
Header = NewType("Header", str)

#: A file-like object representing the request input data.
RequestInput = NewType("RequestInput", BinaryIO)

#: A bytestring representing the request input data.
RequestBody = NewType("RequestBody", bytes)

#: Parsed request data.
RequestData = NewType("RequestData", Dict[str, Any])


def extract_optional_annotation(annotation: Any) -> Tuple[bool, Any]:
    """Returns a tuple denoting whether or not the annotation is an
    Optional type and the inner annotation.
    """
    if is_optional_annotation(annotation):
        return True, annotation.__args__[0]
    return False, annotation


def is_optional_annotation(annotation: Any) -> bool:
    """Returns True if the given annotation represents an Optional type.
    """
    try:
        return getattr(annotation, "__origin__", None) is Union and \
            issubclass(annotation.__args__[1], type(None))
    except TypeError:  # pragma: no cover
        return False



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum, auto

from .http import HTTP_301, HTTP_302, HTTP_307, HTTP_308, Response


class RedirectType(Enum):
    TEMPORARY = auto()
    PERMANENT = auto()


#: A mapping from redirect types to HTTP/1.0 status codes.
LEGACY_REDIRECT_CODES = {
    RedirectType.TEMPORARY: HTTP_302,
    RedirectType.PERMANENT: HTTP_301,
}

#: A mapping from redirect types to HTTP/1.1 status codes.  The
#: advantage of these codes is the request method is preserved during
#: the redirect.
MODERN_REDIRECT_CODES = {
    RedirectType.TEMPORARY: HTTP_307,
    RedirectType.PERMANENT: HTTP_308,
}


def redirect(
        target_location: str,
        *,
        redirect_type: RedirectType = RedirectType.TEMPORARY,
        use_modern_codes: bool = True,
) -> Response:
    """Construct an HTTP Response to redirect the client elsewhere.

    Parameters:
      target_location: Where the client should be redirected to.
      redirect_type: PERMANENT or TEMPORARY.
      use_modern_codes: Whether or not to use HTTP/1.1 response codes.
        The advantage to using HTTP/1.1 codes is the request method is
        preserved during redirect, but older clients (IE11 and older)
        might not support them.
    """
    if use_modern_codes:
        status_code = MODERN_REDIRECT_CODES[redirect_type]
    else:
        status_code = LEGACY_REDIRECT_CODES[redirect_type]

    return Response(status_code, headers={
        "Location": target_location
    })



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import hmac
import json
from inspect import Parameter
from time import time
from typing import Any, Callable, Dict, Optional, no_type_check
from uuid import uuid4

from typing_extensions import Protocol

from molten.dependency_injection import DependencyResolver
from molten.http import Cookie, Cookies

#: The name of the key the CookieStore inserts into session objects to
#: represent the expiration time of the session.
COOKIE_EXPIRATION_KEY = "__EXP__"

#: The default if the expiration key is not set on the incoming
#: session object.
DEFAULT_EXPIRATION_TIME = float("inf")


class Session(Dict[str, Any]):
    """Session objects are ordinary dictionaries that are guaranteed
    to be constructed with an "id" key.
    """

    def __init__(self, id: str, **data: Dict[str, Any]) -> None:
        super().__init__(id=id, **data)

    @classmethod
    def empty(cls) -> "Session":
        """Create an empty session with a random id.
        """
        return cls(str(uuid4()))


class SessionStore(Protocol):
    """Protocol for session stores.
    """

    @no_type_check
    def load(self) -> Session:
        """Load a session from the request.

        This method may request components via DI.
        """

    @no_type_check
    def dump(self) -> Cookie:
        """Convert the session to a Cookie and possibly store it
        somewhere (eg. memcached, Redis).

        This method may request components via DI.
        """


class CookieStore:
    """A stateless session store based on cookies.  Sessions are
    converted to JSON and then base64-encoded.  The values are signed
    with a signing key and validated when the sessions are
    subsequently loaded.

    An expiration time is inserted into the session before it's dumped
    so as to provide minimal protection against session replay
    attacks.

    Warning:
      Don't store sensitive information in sessions using this store.
      They are tamper-proof, but users can decode them.
    """

    __slots__ = [
        "signing_key",
        "signing_method",
        "cookie_ttl",
        "cookie_name",
        "cookie_domain",
        "cookie_path",
        "cookie_secure",
    ]

    def __init__(
            self,
            signing_key: bytes, *,
            signing_method: str = "sha256",
            cookie_ttl: int = 86400 * 7,
            cookie_name: str = "__sess__",
            cookie_domain: Optional[str] = None,
            cookie_path: Optional[str] = None,
            cookie_secure: bool = False,
    ) -> None:
        self.signing_key = signing_key
        self.signing_method = signing_method
        self.cookie_ttl = cookie_ttl
        self.cookie_name = cookie_name
        self.cookie_domain = cookie_domain
        self.cookie_path = cookie_path

    def load(self, cookies: Cookies) -> Session:
        cookie = cookies.get(self.cookie_name)
        if cookie is None:
            return Session.empty()

        data, _, signature = cookie.partition(",")
        if not hmac.compare_digest(signature, self.sign(data.encode())):
            return Session.empty()

        # Note: at this point, the data is guaranteed to be valid
        # given that the data is correctly signed.
        session_data = json.loads(base64.urlsafe_b64decode(data))
        session = Session(**session_data)
        if session.get(COOKIE_EXPIRATION_KEY, DEFAULT_EXPIRATION_TIME) <= time():
            return Session.empty()

        return session

    def dump(self, session: Session) -> Cookie:
        session[COOKIE_EXPIRATION_KEY] = expires = time() + self.cookie_ttl
        session_data = base64.urlsafe_b64encode(json.dumps(session).encode())
        signature = self.sign(session_data)
        return Cookie(
            self.cookie_name, f"{session_data.decode()},{signature}",
            domain=self.cookie_domain, path=self.cookie_path,
            expires=expires, http_only=True, same_site="strict",
        )

    def sign(self, value: bytes) -> str:
        return hmac.new(self.signing_key, value, self.signing_method).hexdigest()


class SessionComponent:
    """A component that loads Session objects from the request.

    Parameters:
      store: A session store.
    """

    __slots__ = ["store"]

    is_cacheable = True
    is_singleton = False

    def __init__(self, store: SessionStore) -> None:
        self.store = store

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        return parameter.annotation is Session

    def resolve(self, resolver: DependencyResolver) -> Session:
        return resolver.resolve(self.store.load)()


class SessionMiddleware:
    """A middleware that dumps Session data into the response.

    Parameters:
      store: A session store.
    """

    __slots__ = ["store"]

    def __init__(self, store: SessionStore) -> None:
        self.store = store

    def __call__(self, handler: Callable[..., Any]) -> Callable[..., Any]:
        def middleware(resolver: DependencyResolver) -> Callable[..., Any]:
            response = handler()
            response.set_cookie(resolver.resolve(self.store.dump)())
            return response
        return middleware



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from threading import local
from typing import Any, Callable, Optional
from uuid import uuid4

from molten import Header

STATE = local()


def get_request_id() -> Optional[str]:
    """Retrieves the request id for the current thread.
    """
    return getattr(STATE, "request_id", None)


def set_request_id(request_id: Optional[str]) -> None:
    """Set a request id for the current thread.  If ``request_id`` is
    None, then a random id will be generated.
    """
    if request_id is None:
        request_id = str(uuid4())

    STATE.request_id = request_id


class RequestIdFilter(logging.Filter):
    """Adds the current request id to log records, making it possible
    to log request ids via the standard logging module.

    Example logging configuration::

      import logging.config
      logging.config.dictConfig({
          "version": 1,
          "filters": {
              "request_id": {
                  "()": "molten.contrib.request_id.RequestIdFilter"
              },
          },
          "formatters": {
              "standard": {
                  "format": "%(levelname)-8s [%(asctime)s] [%(request_id)s] %(name)s: %(message)s"
              },
          },
          "handlers": {
              "console": {
                  "level": "DEBUG",
                  "class": "logging.StreamHandler",
                  "filters": ["request_id"],
                  "formatter": "standard",
              },
          },
          "loggers": {
              "myapp": {
                  "handlers": ["console"],
                  "level": "DEBUG",
                  "propagate": False,
              },
          }
      })
    """

    def filter(self, record: Any) -> bool:
        record.request_id = get_request_id()
        return True


class RequestIdMiddleware:
    """Adds an x-request-id to responses containing a unique request
    id value.  If the incoming request has an x-request-id header then
    that value is reused for the response.  This makes it easy to
    trace requests within a microservice architecture.
    """

    def __call__(self, handler: Callable[..., Any]) -> Callable[..., Any]:
        def middleware(x_request_id: Optional[Header]) -> Any:
            set_request_id(x_request_id)
            response = handler()
            response.headers.add("x-request-id", get_request_id())
            return response
        return middleware



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This module is tested using the autobahn testsuite:
# https://github.com/crossbario/autobahn-testsuite

import io
import logging
import socket
import struct
from base64 import b64encode
from concurrent.futures import Future, ThreadPoolExecutor
from hashlib import sha1
from inspect import Parameter
from typing import Any, Callable, Optional, Pattern, Union

from molten import (
    HTTP_400, HTTP_426, BaseApp, DependencyResolver, Environ, HeaderMissing, HTTPError, MoltenError,
    Request, RequestHandled, Response, Route, TestClient
)
from molten.http.headers import Headers, HeadersDict
from molten.http.query_params import ParamsDict, QueryParams

try:
    import gevent
except ImportError:  # pragma: no cover
    raise ImportError("'gevent' package missing. Run 'pip install gevent'.")


LOGGER = logging.getLogger(__name__)

#: The pre-shared key defined in the Websocket spec.
PSK = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"

#: The amount of bytes to request per recv call.
CHUNKSIZE = 16 * 1024

#: The maximum number of bytes text and binary frames can contain.
MAX_MESSAGE_SIZE = 16 * 1024 * 1024

#: The maximum number of bytes data frames can contain.
MAX_DATA_FRAME_PAYLOAD_SIZE = MAX_MESSAGE_SIZE

#: The maximum number of bytes control frames can contain.
MAX_CONTROL_FRAME_PAYLOAD_SIZE = 125

#: Continuation frame.  Only valid if received after non-final text or binary frames.
OP_CONTINUATION = 0x0

#: A frame with a utf-8 encoded payload.  May or may not be final.
OP_TEXT = 0x1

#: A frame containing binary data.  May or may not be final.
OP_BINARY = 0x2

#: A frame signaling that the connection should be closed.  Always final.
OP_CLOSE = 0x8

#: A frame signaling that a PONG frame should be sent to the client.  Always final.
OP_PING = 0x9

#: A heartbeat frame.  Always final.
OP_PONG = 0xA

#: The set of data frame opcodes.
DATA_FRAME_OPCODES = {OP_CONTINUATION, OP_TEXT, OP_BINARY}

#: The set of control frame opcodes.
CONTROL_FRAME_OPCODES = {OP_CLOSE, OP_PING, OP_PONG}

#: The set of all valid opcodes.
OPCODES = DATA_FRAME_OPCODES | CONTROL_FRAME_OPCODES

#: The set of valid close message status codes.
VALID_STATUS_CODES = {1000, 1001, 1002, 1003, 1007, 1008, 1009, 1010, 1011}

#: The set of reserved close message status codes.
RESERVED_STATUS_CODES = {1004, 1005, 1006, 1015}

#: The set of supported versions.
SUPPORTED_VERSIONS = {"7", "8", "13"}

#: The set of supported versions as a string.
SUPPORTED_VERSIONS_STR = ",".join(SUPPORTED_VERSIONS)

#: The payload that is returned as part of the connection upgrade process.
UPGRADE_RESPONSE_TEMPLATE = b"\r\n".join([
    b"HTTP/1.1 101 Switching Protocols",
    b"connection: upgrade",
    b"upgrade: websocket",
    b"server: molten",
    b"sec-websocket-accept: %(websocket_accept)s",
    b"\r\n",
])


class WebsocketError(MoltenError):
    """Base class for errors related to websockets.
    """


class WebsocketProtocolError(WebsocketError):
    """Raised whenever the protocol is violated.
    """


class WebsocketMessageTooLargeError(WebsocketProtocolError):
    """Raised when an incoming message contains too much data.
    """


class WebsocketFrameTooLargeError(WebsocketError):
    """Raised when a frame's payload is too large to be sent in a
    single frame.
    """


class WebsocketClosedError(WebsocketError):
    """Raised when a message is sent to a closed socket.
    """


class _BufferedStream:
    """A buffered IO stream backed by a socket.

    This makes data frame parsing simple and efficient because the
    data frame reader can request data in small byte chunks, whereas
    this will read the data in large chunks from the socket under the
    hood.
    """

    __slots__ = ["buf", "closed", "socket"]

    def __init__(self, socket: socket.socket) -> None:
        self.buf = b""
        self.closed = False
        self.socket = socket

    def read(self, n: int) -> bytes:
        while not self.closed and len(self.buf) < n:
            data = self.socket.recv(CHUNKSIZE)
            if not data:
                self.closed = True
                return self.buf

            self.buf += data

        data, self.buf = self.buf[:n], self.buf[n:]
        return data

    def expect(self, n: int) -> bytes:
        data = self.read(n)
        if len(data) != n:
            raise WebsocketProtocolError("Unexpected EOF while reading from socket.")

        return data

    def write(self, data: bytes) -> None:
        self.socket.sendall(data)

    def close(self) -> None:
        self.socket.shutdown(True)
        self.socket.close()


class _DataFrameHeader:
    __slots__ = ["fin", "flags", "opcode", "length", "mask"]

    RSV1_MASK = 0x40
    RSV2_MASK = 0x20
    RSV3_MASK = 0x10

    FIN_MASK = MASK_MASK = 0x80
    FLAGS_MASK = RSV1_MASK | RSV2_MASK | RSV3_MASK
    OPCODE_MASK = 0x0F
    LENGTH_MASK = 0x7F

    def __init__(self, fin: bool = False, flags: int = 0, opcode: int = 0, length: int = 0, mask: Optional[bytearray] = None) -> None:  # noqa
        self.fin = fin
        self.flags = flags
        self.opcode = opcode
        self.length = length
        self.mask = mask or bytearray()

    def mask_data(self, data: bytes) -> bytes:
        data_array = bytearray(data)
        for i in range(self.length):
            data_array[i] ^= self.mask[i % 4]

        return bytes(data_array)

    @classmethod
    def from_stream(cls, stream: _BufferedStream) -> "_DataFrameHeader":
        """Read a data frame header from the input stream.
        """
        read = stream.expect
        data = read(2)

        fb, sb = struct.unpack("!BB", data)
        header = cls(
            fb & cls.FIN_MASK == cls.FIN_MASK,
            fb & cls.FLAGS_MASK,
            fb & cls.OPCODE_MASK,
        )

        length = sb & cls.LENGTH_MASK
        if length == 126:
            header.length = struct.unpack("!H", read(2))[0]

        elif length == 127:
            header.length = struct.unpack("!Q", read(8))[0]

        else:
            header.length = length

        if sb & cls.MASK_MASK == cls.MASK_MASK:
            header.mask = bytearray(read(4))

        return header

    def to_stream(self, stream: _BufferedStream) -> None:
        """Write this header to the output stream.
        """
        output = bytearray()

        fb = self.opcode
        if self.fin:
            fb |= self.FIN_MASK

        if self.flags & self.RSV1_MASK == self.RSV1_MASK:
            fb |= self.RSV1_MASK

        if self.flags & self.RSV2_MASK == self.RSV2_MASK:
            fb |= self.RSV2_MASK

        if self.flags & self.RSV3_MASK == self.RSV3_MASK:
            fb |= self.RSV3_MASK

        output.append(fb)

        sb = self.MASK_MASK if self.mask else 0
        if self.length < 126:
            sb |= self.length
            output.append(sb)

        elif self.length <= 0xFFFF:
            sb |= 126
            output.append(sb)
            output.extend(struct.pack("!H", self.length))

        elif self.length <= 0xFFFFFFFFFFFFFFF:
            sb |= 127
            output.append(sb)
            output.extend(struct.pack("!Q", self.length))

        else:
            raise WebsocketFrameTooLargeError(f"{self.length} bytes cannot fit in a single frame.")

        if self.mask:
            output.extend(self.mask)

        stream.write(output)


class _DataFrame:
    __slots__ = ["header", "data"]

    def __init__(self, header: _DataFrameHeader, data: bytes = b"") -> None:
        self.header = header
        self.data = data

    @classmethod
    def from_stream(cls, stream: _BufferedStream) -> "_DataFrame":
        """Read a data frame from an input stream.
        """
        header = _DataFrameHeader.from_stream(stream)
        if header.opcode not in OPCODES:
            raise WebsocketProtocolError(f"Invalid opcode 0x{header.opcode:x}.")

        if header.flags != 0:
            raise WebsocketProtocolError("Reserved flags must not be set.")

        if header.opcode in CONTROL_FRAME_OPCODES:
            max_size = MAX_CONTROL_FRAME_PAYLOAD_SIZE
        else:
            max_size = MAX_DATA_FRAME_PAYLOAD_SIZE

        if header.length > max_size:
            raise WebsocketMessageTooLargeError(f"Payload exceeds {max_size} bytes.")

        data = stream.expect(header.length)
        if header.mask:
            data = header.mask_data(data)

        return cls(header, data)

    def to_stream(self, stream: _BufferedStream) -> None:
        """Write this data frame to the output stream.
        """
        self.header.to_stream(stream)
        if self.header.mask:
            stream.write(self.header.mask_data(self.data))
        else:
            stream.write(self.data)


class Message:
    """A websocket message, composed of one or more data frames.
    """

    __slots__ = ["buf"]

    def __init__(self, message: bytes = b"") -> None:
        self.buf = io.BytesIO(message)

    @classmethod
    def from_frame(cls, frame: _DataFrame) -> "Message":
        message = cls()
        message.buf.write(frame.data)
        return message

    def add_frame(self, frame: _DataFrame) -> None:  # pragma: no cover
        raise NotImplementedError(f"{type(self).__name__} does not implement add_frame()")

    def to_stream(self, stream: _BufferedStream) -> None:
        """Write this message to the output stream.
        """
        output = self.get_output()
        header = _DataFrameHeader(fin=True, opcode=OPCODES_BY_MESSAGE[type(self)], length=len(output))
        frame = _DataFrame(header, output)  # type: ignore
        frame.to_stream(stream)

    def get_data(self) -> bytes:
        """Get this message's data as a bytestring.
        """
        return self.buf.getvalue()

    def get_text(self) -> str:
        """Get this message's contents as text.
        """
        try:
            return self.buf.getvalue().decode("utf-8")
        except UnicodeDecodeError as e:
            raise WebsocketProtocolError("Invalid UTF-8 payload.") from None

    def get_output(self) -> Union[bytes, bytearray, memoryview]:
        """Get this message's output payload.  CloseMessage hooks into
        this to prepend the status code to the payload.
        """
        return self.buf.getbuffer()


class CloseMessage(Message):
    """Received (or sent) when the connection should be closed.  Close
    messages sent by the client are automatically handled by
    receive().

    Attributes:
      code(int): The close status code.
    """

    __slots__ = ["buf", "code"]

    def __init__(self, code: int = 1000, reason: str = "") -> None:
        self.buf = io.BytesIO(reason.encode("utf-8"))
        self.code = code

    @classmethod
    def from_frame(cls, frame: _DataFrame) -> "Message":
        code = 1000
        if frame.data:
            code_data, frame.data = frame.data[:2], frame.data[2:]
            if len(code_data) < 2:
                raise WebsocketProtocolError("Expected status code in close message payload.")

            code = struct.unpack("!H", code_data)[0]
            if code < 1000 or code > 4999:
                raise WebsocketProtocolError(f"Invalid status code {code}.")

            elif code in RESERVED_STATUS_CODES:
                raise WebsocketProtocolError(f"Status code {code} is reserved.")

            elif code < 3000 and code not in VALID_STATUS_CODES:
                raise WebsocketProtocolError(f"Invalid status code {code}.")

        message = cls()
        message.code = code
        message.buf.write(frame.data)
        return message

    def get_output(self) -> Union[bytes, bytearray, memoryview]:
        return struct.pack("!H", self.code) + self.buf.getvalue()


class BinaryMessage(Message):
    """A message containing binary data.
    """

    def add_frame(self, frame: _DataFrame) -> None:
        if len(self.buf.getbuffer()) + len(frame.data) > MAX_MESSAGE_SIZE:
            raise WebsocketProtocolError(f"Message exceeds {MAX_MESSAGE_SIZE} bytes.")

        self.buf.write(frame.data)


class TextMessage(BinaryMessage):
    """A message containing text data.
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message.encode("utf-8"))


class PingMessage(Message):
    """A PING message.  These are automatically handled by receive().
    """


class PongMessage(Message):
    """A PONG message.  These are automatically handled by receive().
    """


#: A mapping from message classes to opcodes.
OPCODES_BY_MESSAGE = {
    CloseMessage: OP_CLOSE,
    BinaryMessage: OP_BINARY,
    TextMessage: OP_TEXT,
    PingMessage: OP_PING,
    PongMessage: OP_PONG,
}


class Websocket:
    """Represents a single websocket connection.  These are used for
    bi-directional communication with a websocket client.

    Websockets are *not* thread-safe.

    Example:
      >>> from molten import annotate
      >>> from molten.contrib.websockets import CloseMessage, Websocket

      >>> @annotate(supports_ws=True)
      ... def echo(sock: Websocket):
      ...     while not sock.closed:
      ...         message = sock.receive()
      ...         if isinstance(message, CloseMessage):
      ...             break
      ...
      ...         sock.send(message)

    Attributes:
      closed(bool): Whether or not this socket has been closed.
    """

    __slots__ = ["closed", "stream"]

    def __init__(self, stream: _BufferedStream) -> None:
        self.closed = False
        self.stream = stream

    def receive(self, *, timeout: Optional[float] = None) -> Optional[Message]:
        """Waits for a message from the client for up to *timeout* seconds.
        """
        if self.closed:
            return None

        with gevent.Timeout(timeout):
            message = None
            while True:
                frame = _DataFrame.from_stream(self.stream)
                if frame.header.opcode == OP_TEXT:
                    if message is not None:
                        raise WebsocketProtocolError("Unexpected text frame.")

                    message = TextMessage.from_frame(frame)

                elif frame.header.opcode == OP_BINARY:
                    if message is not None:
                        raise WebsocketProtocolError("Unexpected binary frame.")

                    message = BinaryMessage.from_frame(frame)

                elif frame.header.opcode == OP_CONTINUATION:
                    if message is None:
                        raise WebsocketProtocolError("Unexpected continuation frame.")

                    message.add_frame(frame)

                elif frame.header.opcode == OP_CLOSE:
                    if not frame.header.fin:
                        raise WebsocketProtocolError("Close frame is not final.")

                    message = CloseMessage.from_frame(frame)
                    self.close(CloseMessage(reason=message.get_text()))
                    return message

                elif frame.header.opcode == OP_PING:
                    if not frame.header.fin:
                        raise WebsocketProtocolError("Ping frame is not final.")

                    self.send(PongMessage(frame.data))
                    continue

                elif frame.header.opcode == OP_PONG:
                    if not frame.header.fin:
                        raise WebsocketProtocolError("Pong frame is not final.")

                    continue

                else:
                    raise WebsocketProtocolError(f"Unexpected frame with opcode 0x{frame.header.opcode:x}.")

                if frame.header.fin:
                    return message

    def send(self, message: Message) -> None:
        """Send a message to the client.
        """
        if self.closed:
            raise WebsocketClosedError("Websocket already closed.")

        message.to_stream(self.stream)

    def close(self, message: Optional[Message] = None) -> None:
        """Close this websocket and send a close message to the client.

        Note:
          This does not close the underlying websocket as it's better
          to let gunicorn handle that by itself.
        """
        try:
            self.send(message or CloseMessage())
            self.stream.close()
        except WebsocketClosedError:
            pass
        finally:
            self.closed = True


class _WebsocketComponent:
    """Resolves websocket objects.  Users of this module don't need to
    worry about providing this to the App object as the middleware
    does it automatically.
    """

    __slots__ = ["websocket"]

    is_cacheable = True
    is_singleton = False

    def __init__(self, websocket: Websocket) -> None:
        self.websocket = websocket

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        return parameter.annotation is Websocket

    def resolve(self) -> Websocket:
        return self.websocket


class WebsocketsMiddleware:
    """A middleware that handles websocket upgrades.

    Warning:
      Please note that this functionality is currently gunicorn-specific
      and it requires the use of async workers in order to function
      correctly.

    Parameters:
      origin_re: An optional regular expression that can be used to
        validate the origin of incoming browser requests.
    """

    __slots__ = ["origin_re"]

    def __init__(self, origin_re: Optional[Pattern[str]] = None) -> None:
        self.origin_re = origin_re

    def handle_exception(self, exception: BaseException, websocket: Websocket) -> None:
        """Called whenever an unhandled exception occurs in middleware
        or a handler.  Overwrite this in a subclass to implement
        custom error handling for websocket handlers.

        If you do overwrite this, don't forget to close the websocket
        connection when necessary.
        """
        LOGGER.exception("Unhandled error from websocket handler.")

        if issubclass(type(exception), WebsocketProtocolError):
            websocket.close(CloseMessage(1002, str(exception)))
        elif issubclass(type(exception), WebsocketFrameTooLargeError):
            websocket.close(CloseMessage(1009, str(exception)))
        else:
            websocket.close(CloseMessage(1011, "Internal server error."))

    def __call__(self, handler: Callable[..., Any]) -> Callable[..., Response]:
        def handle(
                resolver: DependencyResolver,
                request: Request,
                environ: Environ,
                route: Optional[Route],
        ) -> Response:

            if route is None or not getattr(route.handler, "supports_ws", False):
                return handler()

            try:
                connection = request.headers["connection"]
                upgrade = request.headers["upgrade"]
                websocket_key = request.headers["sec-websocket-key"]
                websocket_version = request.headers["sec-websocket-version"]
            except HeaderMissing as e:
                raise HTTPError(HTTP_400, {"errors": {str(e): "this header is required"}})

            try:
                origin = request.headers["origin"]
            except HeaderMissing:
                origin = ""

            if self.origin_re and not self.origin_re.match(origin):
                raise HTTPError(HTTP_400, {"error": "invalid origin"})

            if "upgrade" not in connection.lower() or "websocket" not in upgrade.lower():
                raise HTTPError(HTTP_400, {"error": "invalid upgrade request"})

            if websocket_version not in SUPPORTED_VERSIONS:
                return Response(HTTP_426, headers={"sec-websocket-version": SUPPORTED_VERSIONS_STR})

            # TODO: Implement extension handling.
            # TODO: Implement subprotocol handling.
            stream = _BufferedStream(environ["gunicorn.socket"])
            stream.write(UPGRADE_RESPONSE_TEMPLATE % {
                b"websocket_accept": b64encode(sha1(f"{websocket_key}{PSK}".encode()).digest()),
            })

            websocket = Websocket(stream)
            resolver.add_component(_WebsocketComponent(websocket))

            try:
                handler()
            except Exception as e:
                handle_exception = resolver.resolve(self.handle_exception)
                handle_exception(exception=e)
            finally:
                websocket.close(CloseMessage())

            raise RequestHandled("websocket request was upgraded")
        return handle


class _WebsocketsTestConnection:
    """A proxy context manager for websocket objects.
    """

    __slots__ = ["__future", "__socket"]

    def __init__(self, future: Future, socket: Websocket) -> None:  # type: ignore
        self.__future = future
        self.__socket = socket

    def close(self) -> None:
        try:
            self.__socket.send(CloseMessage())
        except WebsocketClosedError:
            pass
        finally:
            self.__future.result()

    def __enter__(self) -> "_WebsocketsTestConnection":
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.close()

    def __getattr__(self, name: str) -> Any:
        return getattr(self.__socket, name)


class WebsocketsTestClient(TestClient):
    """This is a subclass of the standard test client that adds an
    additional method called :meth:`.connect` that may be used to
    connect to websocket endpoints.

    Example:
      >>> client = WebsocketsTestClient(app)
      >>> with client.connect("/echo") as sock:
      ...   sock.send(TextMessage("hi!"))
      ...   assert sock.receive(timeout=1).get_text() == "hi!"

    Note:
      In order for :meth:`receive's<Websocket.receive>` "timeout"
      parameter to work, you need use gevent to monkeypatch sockets
      before running your tests.
    """

    def __init__(self, app: BaseApp) -> None:
        self.app = app
        self.executor = ThreadPoolExecutor(max_workers=8)

    def connect(
            self,
            path: str,
            headers: Optional[Union[HeadersDict, Headers]] = None,
            params: Optional[Union[ParamsDict, QueryParams]] = None,
            auth: Optional[Callable[[Request], Request]] = None,
    ) -> _WebsocketsTestConnection:
        """Initiate a websocket connection against the application.

        Parameters:
          path: The request path.
          headers: Optional request headers.
          params: Optional query params.
          auth: An optional function that can be used to add auth
            headers to the request.
        """
        headers = headers or Headers()
        headers["connection"] = "upgrade"
        headers["upgrade"] = "websocket"
        headers["sec-websocket-key"] = b64encode(b"a" * 16).decode()
        headers["sec-websocket-version"] = "13"

        client_sock, server_sock = socket.socketpair()

        def prepare_environ(environ: Environ) -> Environ:
            nonlocal client_sock
            environ["gunicorn.socket"] = client_sock
            return environ

        # Execute the websocket handler in a background thread because
        # it may block while waiting on the socket.  Keep a reference
        # to it so we can keep track of exceptions that occur in the
        # handler.
        future = self.executor.submit(
            self.get, path, headers, params, auth=auth,
            prepare_environ=prepare_environ,
        )

        # Delete the client sock so that, if the future completes w/o
        # upgrading the request, it'll get freed (closed) and we're
        # not stuck waiting for a response below.
        del client_sock

        # Consume the upgrade response and make sure it looks right.
        # This is kind of piggy, but it should be fine.
        response_data = b""
        while b"\r\n\r\n" not in response_data and future.running():
            data = server_sock.recv(CHUNKSIZE)
            if not data:
                break

            response_data += data

        expected_response = UPGRADE_RESPONSE_TEMPLATE % {
            b"websocket_accept": b"3SC6TZx4582OZaOogPVxMx5CGS0=",
        }
        if not response_data == expected_response:
            raise ValueError(f"Invalid upgrade response: {response_data}. Did you connect() to a standard endpoint?")

        websocket = Websocket(_BufferedStream(server_sock))
        return _WebsocketsTestConnection(future, websocket)



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .app import App, BaseApp
from .common import annotate
from .dependency_injection import Component, DependencyInjector, DependencyResolver
from .errors import (
    DIError, FieldTooLarge, FieldValidationError, FileTooLarge, HeaderMissing, HTTPError,
    MoltenError, ParamMissing, ParseError, RequestHandled, RequestParserNotAvailable, RouteNotFound,
    RouteParamMissing, TooManyFields, ValidationError
)
from .helpers import RedirectType, redirect
from .http import (
    Cookie, Cookies, Headers, QueryParams, Request, Response, StreamingResponse, UploadedFile
)
from .http.status_codes import *
from .middleware import ResponseRendererMiddleware
from .parsers import JSONParser, MultiPartParser, RequestParser, URLEncodingParser
from .renderers import JSONRenderer, ResponseRenderer
from .router import Include, Route, Router
from .settings import Settings, SettingsComponent
from .testing import TestClient, TestResponse, to_environ
from .typing import (
    Environ, Header, Host, Method, Middleware, Port, QueryParam, QueryString, RequestBody,
    RequestData, RequestInput, Scheme, StartResponse
)
from .validation import (
    Field, Missing, Validator, dump_schema, field, is_schema, load_schema, schema
)

__version__ = "0.5.1"

__all__ = [
    "BaseApp", "App", "Middleware", "annotate",

    # Settings
    "Settings", "SettingsComponent",

    # Router
    "Router", "Route", "Include",

    # WSGI
    "Environ", "StartResponse",

    # HTTP
    "Method", "Scheme", "Host", "Port", "QueryString", "QueryParams", "QueryParam",
    "Headers", "Header", "RequestInput", "RequestBody", "RequestData", "Cookies", "Cookie",
    "UploadedFile", "Request", "Response", "StreamingResponse",

    # Dependency-injection
    "DependencyInjector", "DependencyResolver", "Component",

    # Parsers
    "RequestParser", "JSONParser", "URLEncodingParser", "MultiPartParser",

    # Renderers
    "ResponseRenderer", "JSONRenderer",

    # Middleware
    "ResponseRendererMiddleware",

    # Validation
    "Field", "Missing", "Validator", "field", "schema", "is_schema", "dump_schema", "load_schema",

    # Helpers
    "RedirectType", "redirect",

    # Errors
    "MoltenError", "DIError", "HTTPError", "RouteNotFound", "RouteParamMissing", "RequestHandled",
    "RequestParserNotAvailable", "ParseError", "FieldTooLarge", "FileTooLarge", "TooManyFields",
    "HeaderMissing", "ParamMissing", "ValidationError", "FieldValidationError",

    # Testing
    "TestClient", "TestResponse", "to_environ",

    # Status codes
    # 1xx
    "HTTP_100", "HTTP_101", "HTTP_102",

    # 2xx
    "HTTP_200", "HTTP_201", "HTTP_202", "HTTP_203", "HTTP_204", "HTTP_205", "HTTP_206", "HTTP_207", "HTTP_208",

    # 3xx
    "HTTP_300", "HTTP_301", "HTTP_302", "HTTP_303", "HTTP_304", "HTTP_305", "HTTP_307", "HTTP_308",

    # 4xx
    "HTTP_400", "HTTP_401", "HTTP_402", "HTTP_403", "HTTP_404", "HTTP_405", "HTTP_406", "HTTP_407", "HTTP_408",
    "HTTP_409", "HTTP_410", "HTTP_411", "HTTP_412", "HTTP_413", "HTTP_414", "HTTP_415", "HTTP_416", "HTTP_417",
    "HTTP_418", "HTTP_421", "HTTP_422", "HTTP_423", "HTTP_424", "HTTP_426", "HTTP_428", "HTTP_429", "HTTP_431",
    "HTTP_444", "HTTP_451", "HTTP_499",

    # 5xx
    "HTTP_500", "HTTP_501", "HTTP_502", "HTTP_503", "HTTP_504", "HTTP_505", "HTTP_506", "HTTP_507", "HTTP_508",
    "HTTP_510", "HTTP_511", "HTTP_599",
]



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Callable

from .app import BaseApp
from .errors import HeaderMissing, HTTPError
from .http import HTTP_200, HTTP_406, Request, Response
from .http.headers import HeadersDict


class ResponseRendererMiddleware:
    """A middleware that renders responses.
    """

    def __call__(self, handler: Callable[..., Any]) -> Callable[..., Response]:
        def handle(app: BaseApp, request: Request) -> Response:
            try:
                headers: HeadersDict = {}
                response = handler()
                if isinstance(response, Response):
                    return response

                elif isinstance(response, tuple):
                    if len(response) == 2:
                        status, response = response

                    elif len(response) == 3:
                        status, response, headers = response

                    else:
                        raise RuntimeError("Response tuple must be (status, data) or (status, data, headers).")

                else:
                    status, response = HTTP_200, response

            except HTTPError as e:
                status, response, headers = e.status, e.response, e.headers

            try:
                accept_header = request.headers["accept"]
            except HeaderMissing:
                accept_header = "*/*"

            for mime in accept_header.split(","):
                mime, _, _ = mime.partition(";")

                for renderer in app.renderers:
                    if mime == "*/*" or renderer.can_render_response(mime):
                        response = renderer.render(status, response)
                        response.headers.add_all(headers)
                        return response

            return Response(HTTP_406, content="Not Acceptable")

        return handle
